<?php

namespace FiloBlu\Flow\Model\Indexer;

use Exception;
use Magento\Catalog\Model\Product;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DB\Adapter\AdapterInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Indexer\ActionInterface;
use Magento\Framework\Mview\ActionInterface as MviewActionInterface;
use Magento\Framework\ObjectManagerInterface;
use Magento\Inventory\Model\ResourceModel\StockSourceLink\Collection;
use Magento\Inventory\Model\StockSourceLink;
use Magento\InventoryApi\Api\Data\SourceItemInterface;
use Magento\InventoryConfigurationApi\Api\GetStockItemConfigurationInterface;
use Magento\InventoryConfigurationApi\Exception\SkuIsNotAssignedToStockException;
use Magento\InventoryReservationsApi\Model\GetReservationsQuantityInterface;
use Psr\Log\LoggerInterface;
use Throwable;

/**
 *
 */
class Stock implements ActionInterface, MviewActionInterface
{
    const INDEXER_ID = 'filoblu_flow_stock';
    const XML_PATH_ENABLE = 'filoblu_flow/flow_stock_indexer/enable';

    private $sourceItemRepository;
    /**
     * @var SearchCriteriaBuilder
     */
    private $searchCriteriaBuilder;
    /**
     * @var ResourceConnection
     */
    private $resourceConnection;
    /**
     * @var Product
     */
    private $product;
    /**
     * @var LoggerInterface
     */
    private $logger;
    /**
     * @var GetReservationsQuantityInterface
     */
    private $getReservationsQuantity;
    /**
     * @var GetStockItemConfigurationInterface
     */
    private $getStockItemConfiguration;
    /**
     * @var AdapterInterface
     */
    private $connection;
    /**
     * @var Collection
     */
    private $sourceCollection;
    /**
     * @var  array|null
     */
    private $sourcesToStockId;
    /**
     * @var array|null
     */
    private $stockIdToSources;
    /**
     * @var ScopeConfigInterface
     */
    private $scopeConfig;

    /**
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     * @param ObjectManagerInterface $objectManager
     * @param ResourceConnection $resourceConnection
     * @param Product $product
     * @param LoggerInterface $logger
     * @param ScopeConfigInterface $scopeConfig
     */
    public function __construct(
        SearchCriteriaBuilder  $searchCriteriaBuilder,
        ObjectManagerInterface $objectManager,
        ResourceConnection     $resourceConnection,
        Product                $product,
        LoggerInterface        $logger,
        ScopeConfigInterface   $scopeConfig
    )
    {
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;

        if (interface_exists(\Magento\InventoryApi\Api\SourceItemRepositoryInterface::class)) {
            $this->sourceItemRepository = $objectManager->get(\Magento\InventoryApi\Api\SourceItemRepositoryInterface::class);
        }

        if (class_exists(Collection::class)) {
            $this->sourceCollection = $objectManager->get(Collection::class);
        }

        if (interface_exists(GetReservationsQuantityInterface::class)) {
            $this->getReservationsQuantity = $objectManager->get(GetReservationsQuantityInterface::class);
        }

        if (interface_exists(GetStockItemConfigurationInterface::class)) {
            $this->getStockItemConfiguration = $objectManager->get(GetStockItemConfigurationInterface::class);
        }

        $this->resourceConnection = $resourceConnection;
        $this->product = $product;
        $this->logger = $logger;
        $this->scopeConfig = $scopeConfig;
    }

    /**
     *
     */
    public function executeFull()
    {
        $this->execute(null);
    }

    /**
     * @param int[] $ids
     */
    public function execute($ids)
    {
        if (!$this->sourceItemRepository) {
            return;
        }

        if (!$this->scopeConfig->getValue(self::XML_PATH_ENABLE)) {
            return;
        }

        try {

            if ($ids) {
                $ids = array_unique($ids);
                $this->searchCriteriaBuilder->addFilter('source_item_id', $ids, 'in');
            }

            $result = $this->sourceItemRepository->getList($this->searchCriteriaBuilder->create());

            foreach ($result->getItems() as $sourceItem) {
                //
                // @see \Magento\InventoryConfiguration\Model\GetLegacyStockItem::execute
                // Stock::DEFAULT_STOCK_ID is used until we have proper multi-stock item configuration
                //
                $stockItemConfiguration = $this->getStockItemConfiguration->execute(
                    $sourceItem->getSku(),
                    \Magento\CatalogInventory\Model\Stock::DEFAULT_STOCK_ID

                );

                // We don't need to Manage Stock
                if (!$stockItemConfiguration->isManageStock() || $stockItemConfiguration->getBackorders()) {
                    continue;
                }

                $this->updateLegacyStockFromSourceItem($sourceItem);
            }

        } catch (Exception $e) {
            $this->logger->critical('FiloBlu_Flow_Stock: ' . $e->getMessage());
        } catch (Throwable $t) {
            $this->logger->critical('FiloBlu_Flow_Stock: ' . $t->getMessage());
        }
    }

    /**
     * @param SourceItemInterface $sourceItem
     * @return int
     * @throws LocalizedException
     * @throws SkuIsNotAssignedToStockException
     */
    public function updateLegacyStockFromSourceItem(SourceItemInterface $sourceItem)
    {
        $sourceCode = $sourceItem->getSourceCode();
        $sku = $sourceItem->getSku();
        $stockIds = $this->getStockId($sourceCode);
        $productId = $this->product->getIdBySku($sourceItem->getSku());

        if(!$productId) {
            $this->logger->alert("Product with sku {$sourceItem->getSku()} not longer exists");
            return null;
        }

        foreach ($stockIds as $stockId) {
            $quantity = 0;
            $stockStatus = SourceItemInterface::STATUS_OUT_OF_STOCK;
            $stockQuantity = $this->getQuantity($sku, $stockId);
            // TODO: fix issue calculating reservation on stockId when one source is shared between two stockId
            $reservations = $this->getReservationsQuantity->execute($sourceItem->getSku(), $stockId);
            $minimumQuantity = $this->getStockItemConfiguration->execute($sourceItem->getSku(), $stockId)->getMinQty();

            $realStockStatus = (int)($stockQuantity + $reservations - $minimumQuantity) > 0 ?
                SourceItemInterface::STATUS_IN_STOCK : SourceItemInterface::STATUS_OUT_OF_STOCK;

            if ($realStockStatus != SourceItemInterface::STATUS_OUT_OF_STOCK) {
                $quantity = $stockQuantity;
                $stockStatus = SourceItemInterface::STATUS_IN_STOCK;
            }

            // For default legacy stock
            if ($sourceItem->getSourceCode() === 'default') {
                $this->updateSourceItem($sourceItem, $stockStatus);
                $this->updateStockItem($productId, $quantity, $stockStatus);
                $this->updateStockStatus($productId, $quantity, $stockStatus);
                continue;
            }

            // For sources != default
            // If product in stock is out of stock we force all sources associated to be out of stock

            if ($stockStatus === SourceItemInterface::STATUS_OUT_OF_STOCK) {
                $this->updateStockSourcesStatus($stockId, $sku, SourceItemInterface::STATUS_OUT_OF_STOCK);
            }

            $this->getConnection()->update(
                $this->getConnection()
                    ->getTableName("inventory_stock_$stockId"),
                [
                    'is_salable' => $stockStatus,
                    'quantity'   => $quantity
                ],
                "sku = '$sku' AND (is_salable != $stockStatus OR quantity != $quantity)"
            );
        }

        return $productId;
    }

    /**
     * @param string $sourceCode
     * @return array
     */
    public function getStockId(string $sourceCode): array
    {
        if ($this->sourcesToStockId === null) {
            $this->createLookupTables();
        }

        return $this->sourcesToStockId[$sourceCode];
    }

    /**
     *
     */
    public function createLookupTables()
    {
        $this->sourcesToStockId = [];
        $this->stockIdToSources = [];

        /** @var StockSourceLink $sourcesToStock */
        foreach ($this->sourceCollection->getItems() as $sourcesToStock) {
            $id = $sourcesToStock->getStockId();
            $code = $sourcesToStock->getSourceCode();
            $this->sourcesToStockId[$code][$id] = $id;
            $this->stockIdToSources[$id][$code] = $code;
        }
    }

    /**
     * @param $sku
     * @param $stockId
     * @return float
     */
    public function getQuantity($sku, $stockId): float
    {
        $sources = implode(
            ',',
            array_map(static function ($a) {
                return "'{$a}'";
            }, $this->getSourcesFromStockId($stockId))
        );

        return (float)$this->getConnection()
            ->fetchOne("SELECT COALESCE(SUM(quantity),0) AS `quantity` FROM `inventory_source_item` WHERE `sku` = '{$sku}' AND `source_code` IN ({$sources});");
    }

    /**
     * @param $stockId
     * @return mixed
     */
    public function getSourcesFromStockId($stockId)
    {
        if ($this->stockIdToSources === null) {
            $this->createLookupTables();
        }

        return $this->stockIdToSources[$stockId];
    }

    /**
     * @return AdapterInterface
     */
    public function getConnection(): AdapterInterface
    {
        if ($this->connection === null) {
            $this->connection = $this->resourceConnection->getConnection();
        }

        return $this->connection;
    }

    /**
     * @param SourceItemInterface $sourceItem
     * @param int $isInStock
     */
    public function updateSourceItem(SourceItemInterface $sourceItem, int $isInStock): void
    {
        $connection = $this->getConnection();
        $tableName = $connection->getTableName('inventory_source_item');
        $connection->update($tableName,
            ['status' => $isInStock],
            "sku =  '{$sourceItem->getSku()}' AND source_code = '{$sourceItem->getSourceCode()}' AND status != $isInStock"
        );
    }

    /**
     * @param $productId
     * @param $quantity
     * @param $isInStock
     */
    public function updateStockItem($productId, $quantity, $isInStock): void
    {
        $connection = $this->getConnection();
        $tableName = $connection->getTableName('cataloginventory_stock_item');
        $connection->update($tableName,
            [
                'is_in_stock' => $isInStock,
                'qty'         => $quantity
            ],
            "product_id = $productId AND ( is_in_stock != $isInStock OR qty != $quantity )"
        );
    }

    /**
     * @param int $productId
     * @param int $isInStock
     * @param int $quantity
     */
    public function updateStockStatus(int $productId, int $quantity, int $isInStock)
    {
        $connection = $this->getConnection();
        $tableName = $connection->getTableName('cataloginventory_stock_status');
        $connection->update($tableName,
            [
                'stock_status' => $isInStock,
                'qty'          => $quantity
            ],
            "product_id = $productId AND ( stock_status != $isInStock OR qty != $quantity )"
        );
    }

    /**
     * @param $stockId
     * @param $sku
     * @param $status
     */
    public function updateStockSourcesStatus($stockId, $sku, $status)
    {
        $sources = implode(
            ',',
            array_map(static function ($a) {
                return "'{$a}'";
            }, $this->getSourcesFromStockId($stockId))
        );

        $this->getConnection()->update(
            'inventory_source_item',
            [
                'status' => $status
            ],
            "source_code IN ($sources) AND sku = '$sku' AND status != $status"
        );
    }

    /**
     * @param array $ids
     */
    public function executeList(array $ids)
    {
        $this->execute($ids);
    }

    /**
     * @param int $id
     */
    public function executeRow($id)
    {
        $this->execute([$id]);
    }
}
